################################## NOTES #########################################
# 
##################################################################################
library(ggplot2)
library(plyr)
library(gridExtra)
library(scales)
library(grid)
library(gdata)
library(dplyr)

########### ALL THE PARAMETERS THAT MIGHT NEED TO BE ADJUSTED ###############

withLB = TRUE # If Gurobi is not installed, please set this to FALSE
result_folders = c("ResultsPaper") # folder of the experimental results
data_folder = "DataPaper" # folder of the randomly generated data
numJobs = c(40, 80) # how many jobs exist in the simulation
ylim = 1.13 # limits on y-xis to zoom the figures (set to 1.0185 for c(200,400) jobs)

rhos_str = c("1.5","4.0") # which rho values should be plotted
stdDevs = c("0.5", "0.7", "1.0") # which standard deviation values should be plotted
machines = c(20) # how many machines exist in the simulation
dists = c("lognormal") # name of the distribution
methods = c("BEKP", "LIULU", "LPT", "LB") # all the methods
seeds = 29 # How many seeds to be used starting from 0 (1 less than required number)
epsilons = c(0.05) # all the epsilon values

#############################################################################

getW <- function(numJob, dist, data_folder, seed, stdDev)
{
  data_file = paste("./", data_folder, "/Jobs_NumJobs", numJob, "_", dist, "_seed", seed, "_stdDev", stdDev, ".swf", sep="")
  #print(data_file)
  data_df <- read.csv(data_file, header=FALSE, sep=" ", comment.char=";" )
  W <- sum(data_df[, 2])
  return(W)
}

parse_file <- function(df, result_folder, dist, machine, meth, numJob, rho, eps, seed, stdDev)
{
  result_file = paste("./", result_folder, "/Machines", machine, "_", meth, "_", dist, "_rho", rho, "_eps", eps, 
                      "_NumJobs", numJob, "_seed", seed, "_stdDev", stdDev, ".csv", sep="")
  
  #print(result_file)
  
  result_df <- read.csv(result_file, header=TRUE, sep=";", comment.char="*" )
  
  result_df$dist = rep(dist, numJob)
  result_df$machineNum = rep(machine, numJob)
  result_df$meth = rep(meth, numJob)
  result_df$numJob = rep(numJob, numJob)
  result_df$rho = rep(rho, numJob)
  result_df$eps = rep(eps, numJob)
  result_df$seed = rep(seed, numJob)
  result_df$stdDev = rep(stdDev, numJob)
  
  result_df$dist <- factor(result_df$dist)
  result_df$meth <- factor(result_df$meth)
  result_df$rho <- factor(result_df$rho)
  
  # Extract the comment value from the last line
  last_line <- tail(readLines(result_file), n = 1)
  realSchedulerTime <- strsplit(last_line, " ")[[1]][2]
  
  return(list(as.numeric(realSchedulerTime), result_df)) 
}

findCost <- function(df, m, rho) {
  
  A <- ifelse(nrow(df[df$assignedMachineID != -2, ]) > 0,
              max(df$start_exec_time[df$assignedMachineID != -2] + df$runningTime[df$assignedMachineID != -2]) * m, 0)
  
  RejCost <- ifelse(nrow(df[df$assignedMachineID == -2, ]) > 0,
                    sum(df$runningTime[df$assignedMachineID == -2]) * rho, 0)
  
  return(A + RejCost)
}

total_df <- data.frame()
costs_df <- data.frame()
df <- data.frame()
for (result_folder in result_folders) {
  for (dist in dists) {
    for (numJob in numJobs) {
      for (seed in 0:seeds) {
        for (stdDev in stdDevs) {
          W = getW(numJob, dist, data_folder, seed, stdDev)
          for (rho in rhos_str) {
            #cost_rejectAll = as.numeric(rho) * W
            for (machine in machines) {
              #print(paste("W:", W))
              for (eps in epsilons) {
                if (withLB) {
                  resultLB <- parse_file(df, result_folder, dist, machine, "LB", numJob, rho , eps, seed, stdDev)
                  
                  # in the code the cost is given in place of scheduling time for the LB method!
                  cost_LB <- resultLB[[1]] 
                  dfLB <- resultLB[[2]]
                  
                  # assert if the data file and result file are correct
                  stopifnot(W == sum(dfLB$runningTime))
                }
                else
                  cost_LB <- W
                
                for (meth in methods) {
                  if (meth == "LB") next
                  result <- parse_file(df, result_folder, dist, machine, meth, numJob, rho , eps, seed, stdDev)
                  MethTimeElapsed <- result[[1]]
                  df <- result[[2]]
                  
                  # assert if the data file and result file are correct
                  stopifnot(W == sum(df$runningTime))
                  
                  cost_df = findCost(df, machine, as.numeric(rho))
                  
                  cost_df <- data.frame(cost=cost_df, rel_cost_LB=cost_df/cost_LB, machine = machine, 
                                        meth = meth, numJob = numJob, rho = as.numeric(rho), 
                                        eps = eps, LowerBound = sum(df$runningTime), MethTimeElapsed = MethTimeElapsed, seed = seed, stdDev = stdDev)
                  costs_df <- rbind (cost_df, costs_df)
                }
              }
            }
          }
        }
      }
    }
  }
}

costs_df$stdDev <- factor(costs_df$stdDev, levels = unique(sort(costs_df$stdDev, decreasing = TRUE)))

generate_CostTime_vs_StdDev <- function(plot_df, additional_file_name, withLB, ylim)
{
  additional_title = " " 
  
  p <- ggplot(plot_df, aes(x = stdDev, y = rel_cost_LB, fill = meth))  
  p <- p + geom_boxplot()
  p <- p + facet_grid(numJob ~ rho, scales = "free", 
                      labeller = labeller(rho = function(variable) paste(expression(ρ), "=", variable), 
                                          numJob = function(variable) paste(variable, " jobs")))
  if (withLB)
    p <- p + ylab( "Relative Cost wrt Lower Bound" )
  else 
    p <- p + ylab( "Cost wrt W" )
  p <- p + scale_fill_manual(values = alpha(c("cyan","tomato", "green4", "cyan"), 1))
  p <- p + scale_color_manual(values = alpha(c("cyan","tomato", "green4", "cyan"), 1))
  p <- p + ggtitle(paste(additional_title))
  p <- p + xlab(expression(σ)) 
  p <- p + geom_hline(yintercept = 1, linetype = "dashed", color = "blue") # Add horizontal line
  if (withLB)
    p <- p + coord_cartesian(ylim = c(1, ylim)) 
  p <- p + theme( panel.background = element_rect( fill="white", colour="black"),
                  panel.grid.major = element_line( colour="gray", size=.1 ),
                  legend.position="bottom",
                  legend.title = element_blank(),
                  text = element_text(size = 32),
                  plot.title = element_text(size = 32, face = "bold"),
                  axis.text.x = element_text(angle = 90, hjust = 1, vjust=.5))
  
  ggsave(paste("./StdDevVSCost_SchWithRejection", additional_file_name, ".pdf", sep=""), 
         plot=p, width=10, height=11, device=cairo_pdf)
  
}

#summary(costs_df)

for (numMachine in machines) {
  for (eps in epsilons) {
    generate_CostTime_vs_StdDev( costs_df[costs_df$machine == numMachine & costs_df$eps == eps,], 
                                 paste("_numMachine", numMachine, "_eps", eps, sep=""), withLB, ylim )
  }
}