import math
try:
    from gurobipy import * # For using the Gurobi solver
except ImportError:
    pass
from simulator.solution import Solution
from simulator.set import Set
import bisect

"""Scheduling algorithms module.
New algorithms are added here.

Notes
-----

These parameters are read but NEVER modified by the scheduler. 
It will return a decision, it will NOT actually schedule jobs.

"""

""" This algorithm OPTIMALLY schedules the tasks in a batch 
    way where all tasks are arrived at time 0. 

    Parameters
    ----------
    rho : Float
        Penalty coefficient

    epsilon : Float
        Precision to iterate through possible makespan values

    jobs : List of Job objects
        The jobs to be assigned

    m : Integer
       Number of homogeneous machines in the cluster

    n : Integer
       Global job count in the cluster

    clock : Float (NOT USED IN THIS FUNCTION)
        Specifies the scheduling time of the jobs
        (As all jobs are scheduled at time 0, it is useless 
        for now)
    
    debug : Bool
        If debug mode is ON or NOT

    Returns
    -------
    schedule : Schedule object

    Notes
    -----
        Uses the gurobi linear solver
    """
def ILP(rho, epsilon, jobs, m, n, clock, debug):

    #----------------------------------------------------------
    # GUROBI SOLVER
    #----------------------------------------------------------

    model = Model(name="Paper Solution")
    #model.setParam("TimeLimit", 20.0)
    model.setParam('FeasibilityTol', 1e-6)  # Set feasibility tolerance
    model.setParam('OptimalityTol', 1e-7)  # Set optimality tolerance
    model.setParam('IntFeasTol', 1e-7)

    #----------------------------------------------------------
    # Initialize decision variables
    #----------------------------------------------------------

    #----------------------------------------------------------
    # Decision variable y_ji = 1 if j is 
    # assigned on machine i;
    # otherwise 0
    #----------------------------------------------------------

    y = {}
    for j in range(0,n):
        for i in range(0,m):
            y[j, i] = model.addVar(vtype=GRB.BINARY, name=f"y_{j}_{i}")

    makespan = model.addVar(vtype=GRB.CONTINUOUS, name="max_makespan")

    #----------------------------------------------------------
    # Add constraints
    #----------------------------------------------------------

    # Each job can be assigned to at most one machine
    for j in range(n):
        model.addConstr(quicksum(y[j, i] for i in range(0,m)) <= 1)

    # makespan >= the max occupation time of any machine
    for i in range(m): 
        xrowsum = LinExpr()
        for j in range(n):
            xrowsum += y[j, i] * jobs[j].run
        model.addConstr(makespan >= xrowsum)

    #----------------------------------------------------------
    # Objective function
    #----------------------------------------------------------
        
    rejected_cost = quicksum( rho * jobs[j].run *
                             (1 - quicksum(y[j, i] for i in range(m))) 
                             for j in range(n) )
    
    model.setObjective(makespan * m + rejected_cost, GRB.MINIMIZE)

    #----------------------------------------------------------
    # Solve the problem
    #----------------------------------------------------------

    model.optimize()

    #----------------------------------------------------------
    # Print and return the solution
    #----------------------------------------------------------

    if debug:
        for v in model.getVars():
            print(v)

    X = Solution(m, n)
    for j in range(n):
        for i in range(m):
            if (y[j,i].X):
                X.assignJob(jobs[j], i)
    
    return X

""" This algorithm can only serve as a lower bound on the cost.
    Not a real scheduling algorithm. 

    Parameters
    ----------
    rho : Float
        Penalty coefficient

    epsilon : Float
        Precision to iterate through possible makespan values

    jobs : List of Job objects
        The jobs to be assigned

    m : Integer
       Number of homogeneous machines in the cluster

    n : Integer
       Global job count in the cluster

    clock : Float (NOT USED IN THIS FUNCTION)
        Specifies the scheduling time of the jobs
        (As all jobs are scheduled at time 0, it is useless 
        for now)
    
    debug : Bool
        If debug mode is ON or NOT

    Returns
    -------
    schedule : Schedule object

    Notes
    -----
        Returned solution assumes all accpeted jobs are assigned 
        on machine 0.
    """
def LB(rho, epsilon, jobs, m, n, clock, debug):

    #----------------------------------------------------------
    # GUROBI SOLVER
    #----------------------------------------------------------

    model = Model(name="Paper Solution")
    #model.setParam("TimeLimit", 20.0)
    model.setParam('FeasibilityTol', 1e-8)  # Set feasibility tolerance
    model.setParam('OptimalityTol', 1e-9)  # Set optimality tolerance
    model.setParam('IntFeasTol', 1e-9)

    #----------------------------------------------------------
    # Initialize decision variables
    #----------------------------------------------------------

    #----------------------------------------------------------
    # Decision variable y_j = 1 if App_j is 
    # accepted on node a machine;
    # 0 if rejected
    #----------------------------------------------------------

    y = {}
    for j in range(0,n):
        y[j] = model.addVar(vtype=GRB.BINARY, name=f"y_{j}")

    makespan = model.addVar(vtype=GRB.CONTINUOUS, name="makespan")

    #----------------------------------------------------------
    # Add constraints
    #----------------------------------------------------------

    model.addConstr(makespan >= quicksum((y[j] * jobs[j].run)/m for j in range(n)))

    for j in range(n):
        model.addConstr(makespan >= y[j] * jobs[j].run)

    #----------------------------------------------------------
    # Objective function
    #----------------------------------------------------------
        
    rejected_cost = quicksum( rho * jobs[j].run * (1 - y[j]) 
                              for j in range(n) )
    
    model.setObjective(makespan * m + rejected_cost, GRB.MINIMIZE)

    #----------------------------------------------------------
    # Solve the problem
    #----------------------------------------------------------

    model.optimize()

    #----------------------------------------------------------
    # Print and return the solution
    #----------------------------------------------------------

    if debug:
        for v in model.getVars():
            print(v)

    X = Solution(m, n)
    for j in range(n):
        if (y[j].X):
            X.assignJob(jobs[j], 0)
    X.extra_info = model.getObjective().getValue() 
    
    return X

####### FUNCTIONS RELATED TO BEKP ##############################

""" This algorithm assign jobs from possible set combinations

    Parameters
    ----------
    combs : List of single, tuple or triplet set combinations

    numMachines : integer 
       Total number of machine assignments to make with given 
       combinations

    X : Solution object
        solution to be updated

    machineStart : Integer
        Specifies which node id to start assigning
        (Can be useful when we plot the results)

    debug : bool
        If debug mode is ON or not

    Returns
    -------

    Notes
    -----

    """
def AssignFrom(combs, numMachines, X, machineStart, debug):

    l_i = numMachines
    whichNode = machineStart
        
    for comb in combs: 
        while l_i > 0:
            curr_jobs = getLargestJobs(comb, debug)

            # if there is at least one set in the 
            # combination without available jobs
            if len(curr_jobs) == 0:
                break

            X.assignSeveralJobsToOneMachine(curr_jobs, 
                                            whichNode)
            
            whichNode +=  1 
            l_i -= 1
            
    if l_i > 0: 
        raise Exception("l_i is still greater than 0")
    return 

""" This algorithm can find all possible solutions for a given 
makespan and returns the one with the max area covering one of 
long jobs

    Parameters
    ----------
    jobs : List of Job objects
        The jobs to be assigned

    m : Integer 
       Number of nodes

    T : float
        Makespan
    
    debug : bool
        If debug mode is ON or not

    Returns
    -------
    schedule : Schedule object

    Notes
    -----

    """
def FillMaxArea(jobs, m, T, debug):
    n = len(jobs)
    X = Solution(m, n)
    maxArea = 0

    G, N_1, N_2, N_3, P = generateSubsets(jobs, T)

    if debug:
        print(f'FillMaxArea(C:{T}) --- ' + 
              f'G: {[job.index for job in G.jobs]}, ' + 
              f'N_1: {[job.index for job in N_1.jobs]}, ' + 
              f'N_2: {[job.index for job in N_2.jobs]}, ' + 
              f'N_3: {[job.index for job in N_3.jobs]}, ' + 
              f'P: {[job.index for job in P.jobs]}')

    for j in generateQuadruplets(m, n):

        G.resetCounter()
        N_1.resetCounter()
        N_2.resetCounter()
        N_3.resetCounter()

        if debug:
            print(f'Quadruplet j: {j}')

        X_j = Solution(m, n)

        whichNode = j[0]
        try:
            AssignFrom([(G,), (N_1,), (N_2,), (N_3,)], 
                       j[1], X_j, whichNode, debug)	
        except Exception:
            continue

        if debug:
            print(X_j)

        whichNode += j[1]
        try:
            AssignFrom([(N_2,N_1), (N_3,N_1), (N_2,N_2), 
                        (N_3,N_2), (N_3,N_3)], 
                       j[2], X_j, whichNode, debug)
        except Exception:
            continue

        if debug:
            print(X_j)

        whichNode += j[2]
        try:
            AssignFrom([(N_3,N_3,N_2), (N_3,N_3,N_3)], 
                       j[3], X_j, whichNode, debug)
        except Exception:
            continue

        if debug:
            print(X_j)

        # NON-LAZY version of BEKP
        X_j.assignGreedy(P, (5/4)*T)

        if X_j.getArea() > maxArea:             
            maxArea = X_j.getArea()
            X = X_j

            if debug:
                print(f'++++ X_j: {X_j.getX()} ' + 
                      f'B_j: {X_j.getJobStartTimes()} ' + 
                  f'maxArea: {X_j.getArea()} ' +  
                  f'GlobX: {X.getX()} ' +
                  f'GlobB: {X.getJobStartTimes()} ' + 
                  f'GlobMachineStartTimes: {X.getMachineStartTimes()} ' +
                  f'globMakespan: {X.getMakespan()} ')
	
    # LAZY version of BEKP
    #X.assignGreedy(P, (5/4)*T)

    return(X)

""" This algorithm can schedule the tasks in a batch 
    way where all tasks are arrived at time 0. 

    Parameters
    ----------
    rho : Float
        Penalty coefficient

    epsilon : Float
        Precision to iterate through possible makespan values

    jobs : List of Job objects
        The jobs to be assigned

    m : Integer
       Number of homogeneous machines in the cluster

    n : Integer
       Global job count in the cluster

    clock : Float (NOT USED IN THIS FUNCTION)
        Specifies the scheduling time of the jobs
        (As all jobs are scheduled at time 0, it is useless 
        for now)
    
    debug : Bool
        If debug mode is ON or NOT

    Returns
    -------
    schedule : Schedule object

    Notes
    -----
        In the data generation file, the list of job objects is 
        already soreted in non-increasing order of their runing 
        times. We keep the sorting function just in case for 
        future change in generation file
    """
def BEKP(rho, epsilon, jobs, m, n, clock, debug):
    W = sum(job.run for job in jobs)

    # Default solution is where we reject all jobs (X=X_0)
    X = Solution(m, n)

    # Sort the list of job objects in non-increasing order of 
    # their runing times
    sorted_jobs = sorted(jobs, key=lambda x: x.run, 
                         reverse=True)

    # Compute H, L, k
    H = (4 * W * rho)/(5*m) 
    L = (W * rho) / (5*m*(rho-1))

    H_L_ratio = H/L
    k = math.ceil(math.log(H_L_ratio) / math.log(1 + epsilon))

    if debug:
        print(f'eps: {epsilon} ' + f'rho: {rho} ' + 
              f'm: {m} ' + f'n: {n} ' + 'H: ' + f'{H} ' + 
              f'L: {L} ' + f'W: {W} ' + f'k: {k} ')

    C_i = L
    for i in range(k+1):

        X_i = Solution(m, n)
        X_i = FillMaxArea(sorted_jobs, m, C_i, debug)

        if debug:
            print(f'BEKP iteration i: {i} for C_i: {C_i} ' + 
                  'A: ' + f' {X_i.getArea()} ' + 
                  f'R: {W-X_i.getArea()} ' + 
                  f'makespan: {X_i.getMakespan()} ' + 
                  f'cost: {X_i.getCost(W, m, rho)} ' + 
                  f'GlobCost: {X_i.getCost(W, m, rho)} ' +
                  f'assigned nodes: {[i for i in X_i.assignedMachineIDs]}')

        if (X_i.getCost(W, m, rho) < X.getCost(W, m, rho)):
            X = X_i

        C_i *= (1 + epsilon)
    
    assert(C_i >= H)

    if debug:
        print(f'Result::: ' + 
                  'A: ' + f' {X.getArea()} ' + 
                  f'R: {W-X.getArea()} ' + 
                  f'makespan: {X.getMakespan()} ' + 
                  f'cost: {X.getCost(W, m, rho)} ' + 
                  f'GlobCost: {X.getCost(W, m, rho)} '+
                  f'Assigned nodes: {[i for i in X.assignedMachineIDs]}')
        
    return (X)


""" This algorithm can schedule the tasks in a batch 
    way where all tasks are arrived at time 0. 

    Parameters
    ----------

    jobs : List of Job objects
        The jobs to be assigned

    C : Float
       Target makespan

    Returns
    -------
    All the sets as separate Set objects, where the jobs are 
    sorted by decreasing job duration within each set

    Notes
    -----
    This function assumes that the jobs list is sorted by 
    decreasing job duration

    """
def generateSubsets(jobs, C):

    G = Set()
    N_1 = Set()
    N_2 = Set()
    N_3 = Set()
    P = Set()

    for j in jobs:
        if (j.run <= 5*C/4) and (j.run > 3*C/4):
            G.addJob(j)
        elif (j.run <= 3*C/4) and (j.run > C/2):
            N_1.addJob(j)
        elif (j.run <= C/2) and (j.run > 3*C/8):
            N_2.addJob(j)
        elif (j.run <= 3*C/8) and (j.run > C/4):
            N_3.addJob(j)
        elif (j.run <= C/4):
            P.addJob(j)

    return (G, N_1, N_2, N_3, P)

""" Generator funtion to iterate through all possible 
    quadruplets 

    Parameters
    ----------

    m : Integer
        Number of machines

    n : Integer
        Number of jobs

    Yields
    -------
    All possible quadruplets 

    Notes
    -----

    """
def generateQuadruplets(m, n):   
    for i in range(0, m + 1):
        for j in range(0, m + 1 - i):
            for k in range(0, m + 1 - i - j):
                l = m - i - j - k
                assert(l >= 0)
                if j + (2*k) + (3*l) <= n:
                    yield (i, j, k, l)


""" Get the largest jobs of the sets within the 
    given set combination

    Parameters
    ----------

    comb : List of set combinations

    debug : Bool
        If debug mode is ON or NOT

    Returns
    -------
    List of the largest jobs of the sets within the 
    given set combination

    Notes
    -----

    """
def getLargestJobs(comb, debug):
    curr_jobs = [] 
    
    if debug:
        print("getLargestJobs(comb): ")

    for set, i in zip(comb, range(len(comb))):
        if debug:
            print("set:", 
                  f'{[job.index for job in set.jobs]}',
                  ' counter: ', set.counter, 
                  ' len(set.jobs): ', len(set.jobs))

        if set.counter < len(set.jobs):
            curr_jobs.append(set.jobs[set.counter])
            set.increaseCounter()

            if debug:
                print('curr_set.job.index: ', 
                      set.jobs[set.counter-1].index,  
                      ' curr_set.job.run:', 
                      set.jobs[set.counter-1].run, 
                      ' counter: ', set.counter)
                
        else:
            if debug:
                print('Set in index i: ', i, 'is empty')
                
            for ii in range(i):
                comb[ii].decreaseCounter()
                if debug:
                    print(f'Thus, decrasing the counter of set[{ii}] to: ', 
                          set.counter)
            return ([])
        
    return (curr_jobs)

####### FUNCTIONS RELATED TO LIU ET AL. ##################

""" This algorithm assign jobs with the method in "New 
approximation algorithms for machine scheduling with 
rejection on single and parallel machine" by Liu and Lu 
(uses "Algorithm A 1" there).

    Parameters
    ----------
    rho : Float
        Penalty coefficient

    epsilon : Float (NOT USED IN THIS FUNCTION)
        Precision to iterate through possible makespan values

    jobs : List of Job objects
        The jobs to be assigned

    m : Integer
       Number of homogeneous machines in the cluster
   
    n : Integer
       Global job count in the cluster

    clock : Float (NOT USED IN THIS FUNCTION)
        Specifies the scheduling time of the jobs
        (As all jobs are scheduled at time 0, it is useless 
        for now)
    
    debug : Bool
        If debug mode is ON or NOT

    Returns
    -------
    schedule : Schedule object

    Notes
    -----
        In the data generation file, the list of job objects is 
        already sorted in non-increasing order of their runing 
        times. We keep the sorting function just in case for 
        future change in generation file

        IMPORTANT:
        In this function it is important that the job indices 
        are the same as the jobIds after being sorted in 
        non-increasing order of their runing times (i.e.) 
        sorting operation does not change their indices
    """

def LIULU(rho, epsilon, jobs, m, n, clock, debug):

    # Liu uses Z^Liu = C_max + x R as cost funct. 
    # Thus for a fair comparison, we set x = rho/m
    rho_liu = rho/m

    # Default solution is where we reject all jobs (X=X_0)
    X = Solution(m, n) # step 4 in pdf (=sigma_0 in pdf)
    cost = Z(X, jobs, rho_liu)

    # Sort the list of job objects in non-increasing order of 
    # their runing times (generated data already satisfies this!)
    sorted_jobs = sorted(jobs, 
                         key=lambda x: x.run, 
                         reverse=True)

    if debug:
        print('STARTING LIULU ALGO with -- ' +
              f'eps: {epsilon} ' + f'rho: {rho} ' + 
              f'rho_liu: {rho_liu} ' + f'm: {m} ' + 
              f'n: {n}, cost: {cost*m} ' + 
              f'{[job.index for job in sorted_jobs]}' +
              f'{[job.run for job in sorted_jobs]}')

    for k in range(n):
        p_k = sorted_jobs[k].run

        U_k = [job for job in sorted_jobs if job.run <= p_k]
        
        # Assuming that rho >= 1
        A_k = [job for job in U_k 
               if job.index != k]
        #A_k = [job for job in U_k 
        #       if job.run/m <= rho_liu*job.run 
        #       and job.index != k]

        if debug:
            print(f'A_k indices: {[i.index for i in A_k]}')
        
        index_to_insert = bisect.bisect_left([job.index 
                                              for job in A_k], 
                                              sorted_jobs[k].index)
        
        # TODO : Actually as rho is fixed, A_BAR_k = U_k
        A_BAR_k = A_k[:]
        A_BAR_k.insert(index_to_insert, sorted_jobs[k])
        #R_k = [job for job in U_k if job not in A_BAR_k]

        if debug:
            print(f'k: {k}, p_k : {p_k}, ' +
                  f'U_k indices: {[i.index for i in U_k]}, ' +
                  f'U_k runtimes: {[i.run for i in U_k]}, ' +
                  f'A_k indices: {[i.index for i in A_k]}, ' +
                  f'A_k runtimes: {[i.run for i in A_k]}, ' +
                  f'A_BAR_k indices: {[i.index for i in A_BAR_k]}, ' +
                  f'A_BAR_k runtimes: {[i.run for i in A_BAR_k]} ' + 
                  f'M(A_BAR_k,m): {M(A_BAR_k,m)}, (0.5 + 1/(2*m)) * p_k: {(0.5 + 1/(2*m)) * p_k}') 

        if (M(A_BAR_k,m) < (0.5 + 1/(2*m)) * p_k): # step 2
            # A_k U R_k = U_k \ J_k
            AR_k = [job for job in U_k if job.index != k]
            criterion =  (0.5 - 1/(2*m)) * p_k
            u = find_u(AR_k, criterion, m)

            if debug:
                print(f'Step 2: ' +
                  f'AR_k indices: {[i.index for i in AR_k]}, ' +
                  f'AR_k runtimes: {[i.run for i in AR_k]}, ' +
                  f'criterion: {criterion}, u: {u}') 

            # solution for A_2 and A_1
            for i in range(2):
                A_i = AR_k[:u-i]
                A_i.append(sorted_jobs[k])

                if debug:
                    if (A_i):
                        print(f'A_i[:u-i]: {[i.index for i in A_i[:u-i]]}')
                
                X_k = LPT(rho_liu, epsilon, 
                          A_i, 
                          m, n, clock, debug)
                cost_k = Z(X_k, jobs, rho_liu)
                if (cost_k < cost):
                    if debug:
                        print(f'+ Cost was {cost*m}. Updated global soln with values:')
                    cost = cost_k
                    X = X_k

                if debug:
                    print(f'Solution A_1 or A_2 of Step 2: cost : {Z(X_k, jobs, rho_liu)*m}, ' +
                        f'assigned nodes: {[i for i in X_k.assignedMachineIDs]}, ' +
                        f'start times: {[i for i in X_k.jobStartTimes]}, ' + 
                        f'mach. occupations: {[i for i in X_k.machineStartTimes]} ') 

        else: # step 3
            T_k = [job.run for job in U_k 
                   if job.run < p_k and job.run > p_k/2]
            T_k.append(p_k/2)

            if debug:
                print(f'Step 3:' +
                  f'T_k values: {[i for i in T_k]} ') 

            for t in T_k:
                B_kt = [job for job in A_k if job.run > t]

                if len(B_kt) <= m-1:
                    D_kt = []
                else:
                    D_kt = B_kt[m - 1:]

                    D_kt = sorted(D_kt, key=lambda x: x.run, 
                         reverse=True)

                    # D_k is sorted in non-increasing order
                    assert all(D_kt[i].run >= D_kt[i + 1].run 
                               for i in range(len(D_kt) - 1))
                
                if debug:
                    print(f't: {t}, ' +
                        f'B_kt indices: {[i.index for i in B_kt]}, ' +
                        f'B_kt runtimes: {[i.run for i in B_kt]}, ' +
                        f'D_kt indices: {[i.index for i in D_kt]}, ' +
                        f'D_kt runtimes: {[i.run for i in D_kt]}, ') 
                    
                # solution for A_kt
                X_k = LPT(rho_liu, epsilon, 
                          [job for job in A_BAR_k if job not in D_kt], 
                          m, n, clock, debug)
                cost_k = Z(X_k, jobs, rho_liu)
                if (cost_k < cost):
                    if debug:
                        print(f'+ Cost was {cost*m}. Updated global soln with values:')
                    cost = cost_k
                    X = X_k

                if debug:
                    print(f'Solution of t of Step 3: cost : {Z(X_k, jobs, rho_liu)*m}, ' +
                        f'assigned nodes: {[i for i in X_k.assignedMachineIDs]}, ' +
                        f'start times: {[i for i in X_k.jobStartTimes]}, ' + 
                        f'mach. occupations: {[i for i in X_k.machineStartTimes]} ') 

    if debug:
        print(f'Result::: cost : {cost*m}, ' +
                  f'assigned nodes: {[i for i in X.assignedMachineIDs]}, ' +
                  f'start times: {[i for i in X.jobStartTimes]}, ' + 
                  f'mach. occupations: {[i for i in X.machineStartTimes]} ') 
        
    return (X)

""" This algorithm assign jobs with the method in "New 
approximation algorithms for machine scheduling with 
rejection on single and parallel machine" by Liu and Lu 
(uses "Algorithm A 1" there).

    Parameters
    ----------
    rho : Float
        Penalty coefficient

    epsilon : Float (NOT USED IN THIS FUNCTION)
        Precision to iterate through possible makespan values

    jobs : List of Job objects
        The jobs to be assigned

    m : Integer
       Number of homogeneous machines in the cluster

    n : Integer
       Global job count in the cluster

    clock : Float (NOT USED IN THIS FUNCTION)
        Specifies the scheduling time of the jobs
        (As all jobs are scheduled at time 0, it is useless 
        for now)
    
    debug : Bool
        If debug mode is ON or NOT

    Returns
    -------
    schedule : Schedule object

    Notes
    -----
        Here using the n parameter is important 
        as we call this func in BEKP and there we call 
        this function with subset of all jobs so we 
        need the global jobs number
    """

def LPT(rho, epsilon, jobs, m, n, clock, debug):

    X = Solution(m, n)

    # Sort jobs in descending order of processing times
    if jobs:
        sortedJobs = sorted(jobs, key=lambda x: x.run, reverse=True)

        for job in sortedJobs:
            # Find the machine on which the job can start the soonest
            minMachine = min(enumerate(X.machineStartTimes), key=lambda x: x[1])[0]
            
            X.assignJob(job, minMachine)
    
    return X

def find_u(jobs, value, m):
    cumulative_sum = 0
    for i, job in enumerate(jobs):
        cumulative_sum += job.run/m
        if cumulative_sum > value:
            return i
    return len(jobs)

# Sum of the penalties (rho * p_j) of jobs
def W(jobs, rho):
    return rho * sum(job.run for job in jobs)

# Sum of the processing times (p_j) of jobs
def P(jobs):
    return sum(job.run for job in jobs)

# Sum of the loads (p_j/m) of jobs
def M(jobs, m):
    return sum(job.run for job in jobs)/m

# Get the completion time of a job in the curr schedule
def C_j(sigma, job):
    return sigma.getCompletionTimeOfJob(job)

# Cost according to the cost func in Liu's paper
# Note : Multiply this cost with m to get the cost of our 
# paper and call this function with rho/m value
def Z(sigma, jobs, rho):

    #TODO: here we assume that jobs are ordered according to 
    # their indices (=jobIds)
    rejected_jobs = [job for job, machine_id in 
                  zip(jobs, sigma.assignedMachineIDs) 
                  if machine_id == -2]
    #print(f"sigma.getMakespan():{sigma.getMakespan()}, " +
    #      f"W(rejected_jobs, rho):{W(rejected_jobs, rho)}")
    return sigma.getMakespan() + W(rejected_jobs, rho)

def RejectAll(rho, epsilon, jobs, m, n, clock, debug):
    X = Solution(m, n)
    return X
