"""Simulation engine module.

Replays the jobs from the randomly generated dataset 
through the script 'generate_tasks.py'.

It uses a different scheduling algorithm and a simplified,
configurable-sized system.
"""

import heapq
from random import uniform
from numpy import min, max, mean, median, sum
from simulator.job import Job
from simulator.event import Event
from simulator.utils import printable
import simulator.algorithms as algorithms
import time
import os

class Engine:
    """Simulation engine.
    """
    def __init__(self,
                 algorithmName,
                 numMachines,
                 numJobs,
                 seed,
                 stdDev,
                 dataFolder='data', 
                 distribution='uniform',
                 debug=False):
        """Creates the simulation engine.

        Parameters
        ----------
        algorithmName : String
            Name of the scheduling algorithm to use

        numMachines : Int
            Number of machines in the cluster

        numJobs : Int
            Number of jobs in the cluster

        dataFolder : String 
            Name of the folder containing the Jobs info

        distribution : String
            Defines which data file should be used (the one 
            with execution times generated through 'distribution' 
            type of distribution)

        debug : Bool [default=False]
            True if debugging messages should be printed

        """
        self.numMachines = numMachines
        self.debug = debug
        self.jobs = []
        self.events = []
        self.clock = 0 # non-used for now
        self.numJobs = 0
        self.datafolder = dataFolder

        try:  # gets the scheduling function identified by its name
            self.scheduler = getattr(algorithms, algorithmName)
            if self.debug:
                print(f'DEBUG: Set {algorithmName} as the scheduler.')
        except AttributeError:
            print('PANIC! Could not find scheduling algorithm' +
                  f' {algorithmName}. Stopping execution.')
            exit()

        JobInputFile = str('./' + dataFolder + '/Jobs_NumJobs' + 
            str(numJobs) + '_'+ distribution + '_seed' + seed + '_stdDev' 
            + stdDev + '.swf')
                        
        # Reads the input file to populate 'events' and 'jobs'
        print(f'Reading file {JobInputFile} to populate the simulation')
        
        infile = open(JobInputFile, 'r')

        for line in infile:
            if line[0] == ";":  # skips comments
                continue
            parsed = line.split()

            # Assign the corresponding job values in the file
            jobID = int(parsed[0]) 
            run = float(parsed[1])
            assignedMachine = -1
            submission = 0
            
            # Creates the job and adds it to the list
            newJob = Job(jobID, run, assignedMachine, submission)

            self.numJobs += 1  # another job created
            self.jobs.append(newJob)

        # Closes the file
        infile.close()

        print('Finished reading the input file.' +
            f' {self.numJobs} jobs will be scheduled on' +
            f' {numMachines}' +
            ' machines. Ready for simulation.')

        for job in range(len(self.jobs)):
            if self.debug:
                print('job[', self.jobs[job].jobID ,'], submission: ', 
                      self.jobs[job].submission, ', exec_time: ', self.jobs[job].run)
                
            heapq.heappush(self.events, (self.jobs[job].submission, 
                                         Event(True, self.jobs[job])))

    def run(self, seed, stdDev, rho=5, epsilon=0.05, resultFolder = '/beegfs/ekorkmaz/results', distribution = 'uniform'):
        """Simulates the scheduling of the tasks on the cluster.

        rho : Float [default=5]
            rejection cost of job i is (r_i = rho x p_i), 
            where p_i is the exec time of job

        epsilon : Float [default=0.05]
            Precision to itarate through possbile makespan values

        resultFolder : String 
            Path of the folder which will contain the outputs
        
        Returns
        -------
        bool
            True
        """

        print('Starting the simulation.')

        events = self.events
        jobs = self.jobs
        NumJobs = len(jobs)

        fnameJob = self.initializeOutputFile(NumJobs, resultFolder, 
                                             distribution, rho, epsilon, seed, stdDev)

        with open(fnameJob, 'w') as f:
            # write the header of the output csv file
            title = 'jobID;assignedMachineID;runningTime;'
            title += 'start_exec_time' + '\n'
            f.write(title)

            # list of the jobs that were submitted but not executed yet
            queue = []
            
            # As we know that all jobs in "events" are new job submission 
            # events at time 0, we do not check the event types or submission 
            # times, and put all into our queue
            while len(events) > 0:
                eventClock, newEvent = heapq.heappop(events)      
                queue.append(newEvent.jobInfo)

            print('Start execution of jobs - NumJobs: ',
                    f'{NumJobs} ',
                    f' scheduler: ', self.scheduler)    

            # Checks with the scheduler if there are any
            # jobs that it is able to be scheduled right now
            if len(queue) > 0: # if there are jobs in the ready queue
                    if self.debug:
                        print('Jobs in the queue to schedule:' +
                            f'{printable(queue)} ' + f' clock: {self.clock}')
                        
                    start = time.time()
                    solution = self.scheduler(rho, epsilon, queue, self.numMachines, NumJobs, self.clock, self.debug) 
                    end = time.time()

                    # checks if the scheduler found a suitable job
                    # and updates the global values of simulation
                    if solution.newdecision:

                        if self.debug:
                            for i in range(0,len(solution.assignedMachineIDs)):
                                print('Length event is ', len(events), 
                                    ' and length of queue is ', len(queue), 
                                    '. Job with ID: ', i, 
                                    ' and submission time: ', jobs[i].submission, 
                                    ' is assigned on Machine[', solution.assignedMachineIDs[i], 
                                    '] at time', solution.jobStartTimes[i])
                        
                        # One by one schedule the jobs and update the global variables        
                        for i in range(len(solution.assignedMachineIDs)):

                            # print job info to the file
                            jobInfo = str(i) + ';' + str(solution.assignedMachineIDs[i]) + ';'
                            jobInfo += str(jobs[i].run) + ';' 
                            jobInfo += str(solution.jobStartTimes[i]) + '\n'
                            f.write(jobInfo)

                            if self.debug:
                                print('The job from queue: ', queue[i], 
                                      ' and from self.jobs: ', jobs[i])
                        if solution.extra_info == -2:
                            f.write("* " + str(end-start))
                        else:
                            f.write("* " + str(solution.extra_info))


        return self.clock

    def initializeOutputFile(self, NumJobs, resultFolder, dist, rho, eps, seed, stdDev):

        # If the results folder does not exist, create it
        if not os.path.exists(resultFolder):
            os.makedirs(resultFolder)

        # The output file to print the results related to jobs
        fnameJob = resultFolder + '/Machines'
        fnameJob += str(self.numMachines) 

        if self.scheduler == getattr(algorithms, 'BEKP'):
            fnameJob += '_BEKP'
        if self.scheduler == getattr(algorithms, 'LIULU'):
            fnameJob += '_LIULU'
        if self.scheduler == getattr(algorithms, 'RejectAll'):
            fnameJob += '_RejectAll'
        if self.scheduler == getattr(algorithms, 'LPT'):
            fnameJob += '_LPT'
        if self.scheduler == getattr(algorithms, 'ILP'):
            fnameJob += '_ILP'
        if self.scheduler == getattr(algorithms, 'LB'):
            fnameJob += '_LB'

        fnameJob += '_' + dist + "_rho" + str(rho) + "_eps" + str(eps)
        fnameJob += '_NumJobs' + str(NumJobs) + '_seed' + seed + '_stdDev' 
        fnameJob += stdDev + '.csv'
        
        return fnameJob

    
